@@ -14,7 +14,7 @@ use Time::Piece ();
 has 'mysql';
 has 'no_txn' => sub { 0 };
 
-our $VERSION = '0.27';
+our $VERSION = '0.28';
 
 sub dequeue {
   my ($self, $worker_id, $wait, $options) = @_;
@@ -202,35 +202,35 @@ sub list_jobs {
 
   my $db = $self->mysql->db;
 
-  # Note: The GROUP BY below only needs minion_jobs.id, child_jobs.parent_id,
-  # and parent_jobs.child_id - the additional redundant columns are just
-  # there to satisfy the ONLY_FULL_GROUP_BY requirement in MySQL strict mode.
-  #
-  my $jobs = $db->query(
-    "SELECT
+  my $list_sql = qq{
+    SELECT
       id, args, attempts,
+      state, task, worker,
+      lax, priority, queue, result, retries,
+      children, parents,
       UNIX_TIMESTAMP(created) AS created,
       UNIX_TIMESTAMP(`delayed`) AS `delayed`,
-      UNIX_TIMESTAMP(finished) AS finished, lax, priority,
-      queue, result, UNIX_TIMESTAMP(retried) AS retried, retries,
-      UNIX_TIMESTAMP(started) AS started, state, task,
-      GROUP_CONCAT( child_jobs.child_id ORDER BY child_jobs.child_id SEPARATOR ':' ) AS children,
-      GROUP_CONCAT( parent_jobs.parent_id ORDER BY parent_jobs.parent_id SEPARATOR ':' ) AS parents,
-      worker, UNIX_TIMESTAMP(NOW()) AS time, UNIX_TIMESTAMP(expires) AS expires
-    FROM minion_jobs j
-    LEFT JOIN minion_jobs_depends child_jobs ON j.id=child_jobs.parent_id
-    LEFT JOIN minion_jobs_depends parent_jobs ON j.id=parent_jobs.child_id
-    $where
-    GROUP BY j.id, child_jobs.parent_id, parent_jobs.child_id
-           , j.args, j.attempts, j.created,
-             j.delayed, j.finished, j.lax,
-             j.priority, j.queue, j.result,
-             j.retried, j.retries, j.started,
-             j.state, j.task, j.worker, j.expires
+      UNIX_TIMESTAMP(finished) AS finished,
+      UNIX_TIMESTAMP(retried) AS retried,
+      UNIX_TIMESTAMP(started) AS started,
+      UNIX_TIMESTAMP(NOW()) AS time,
+      UNIX_TIMESTAMP(expires) AS expires
+    FROM minion_jobs
+    JOIN (
+      SELECT id,
+        GROUP_CONCAT( child_jobs.child_id ORDER BY child_jobs.child_id SEPARATOR ':' ) AS children,
+        GROUP_CONCAT( parent_jobs.parent_id ORDER BY parent_jobs.parent_id SEPARATOR ':' ) AS parents
+      FROM minion_jobs j
+      LEFT JOIN minion_jobs_depends child_jobs ON j.id=child_jobs.parent_id
+      LEFT JOIN minion_jobs_depends parent_jobs ON j.id=parent_jobs.child_id
+      $where
+      GROUP BY id
+    ) filtered USING ( id )
     ORDER BY id DESC
-    LIMIT ?
-    OFFSET ?", @params, $limit, $offset,
-  )->hashes;
+    LIMIT ? OFFSET ?
+  };
+
+  my $jobs = $db->query( $list_sql, @params, $limit, $offset )->hashes;
   $jobs->map( _decode_json_fields(qw{ args result }) )
     ->each( sub {
       $_->{children} = [ split /:/, $_->{children} // '' ];
@@ -437,9 +437,10 @@ sub repair {
 
   # Jobs with missing worker (can be retried)
   $db->query(
-    "select id, retries from minion_jobs
+    "select job.id, job.retries from minion_jobs job
+     left join minion_workers worker on job.worker = worker.id
      where state = 'active' and queue != 'minion_foreground'
-       and worker not in (select id from minion_workers)"
+       and worker.id is null"
   )->hashes->each(sub { $self->fail_job(@$_{qw(id retries)}, 'Worker went away') });
 
   # Jobs in queue without workers or not enough workers (cannot be retried and requires admin attention)
@@ -521,45 +522,51 @@ sub retry_job {
 sub stats {
   my $self = shift;
 
-  my $db  = $self->mysql->db;
-  my $all = $db->query('select count(*) from minion_workers')->array->[0];
-  my $sql
-    = "select count(distinct worker) from minion_jobs where state = 'active'";
-  my $active = $db->query($sql)->array->[0];
+  my %stats = map { $_ => 0 } qw(
+    active_workers inactive_workers active_jobs inactive_jobs failed_jobs finished_jobs
+    enqueued_jobs delayed_jobs active_locks uptime
+  );
 
-  #### TODO: odd $a and $b weren't working, or something
-  $sql = 'select state, count(state) from minion_jobs group by 1';
-  my $results
-    = $db->query($sql); # ->reduce(sub { $a->{$b->[0]} = $b->[1]; $a }, {});
+  my $db = $self->mysql->db;
 
-  my $states = {};
-  while (my $next = $results->array) {
-    $states->{$next->[0]} = $next->[1];
+  # Worker statistics
+  my $worker_sql = q{
+    SELECT
+      SUM(job.id IS NOT NULL) AS is_active
+    FROM minion_workers worker
+    LEFT JOIN minion_jobs job
+        ON worker.id = job.worker
+        AND job.state = 'active'
+    GROUP BY worker.id
+  };
+  my $res = $db->query( $worker_sql );
+  while ( my $row = $res->hash ) {
+    $stats{ $row->{is_active} ? 'active_workers' : 'inactive_workers' }++;
   }
 
-  my $uptime = $db->query( "SHOW GLOBAL STATUS LIKE 'Uptime'" )->hash->{Value};
-
-  $sql = q{
+  # Job statistics
+  my $job_sql = q{
     SELECT
-      SUM(CASE WHEN `state` = 'inactive' AND `delayed` > NOW() THEN 1 ELSE 0 END) AS delayed_jobs,
-      COUNT(*) AS enqueued_jobs
-      FROM minion_jobs
-    };
-  %$states = ( %$states, %{ $db->query($sql)->hash } );
-  $states->{active_locks} = $db->query("SELECT COUNT(*) FROM minion_locks WHERE expires > now()")->array->[0];
-
-  return {
-    active_workers   => $active,
-    inactive_workers => $all - $active,
-    active_jobs      => $states->{active} || 0,
-    inactive_jobs    => $states->{inactive} || 0,
-    failed_jobs      => $states->{failed} || 0,
-    finished_jobs    => $states->{finished} || 0,
-    enqueued_jobs    => $states->{enqueued_jobs} || 0,
-    delayed_jobs     => $states->{delayed_jobs} || 0,
-    active_locks     => $states->{active_locks} || 0,
-    uptime           => $uptime || 0,
+        state,
+        COUNT(state) AS jobs,
+        SUM(`delayed` > NOW()) AS `delayed`
+    FROM minion_jobs
+    GROUP BY state
   };
+  $res = $db->query( $job_sql );
+  while ( my $row = $res->hash ) {
+    my $state = $row->{state};
+    $stats{ "${state}_jobs" } += $row->{jobs};
+    $stats{ enqueued_jobs } += $row->{jobs};
+    if ( $state eq 'inactive' ) {
+      $stats{ delayed_jobs } += $row->{delayed};
+    }
+  }
+
+  $stats{active_locks} = $db->query("SELECT COUNT(*) FROM minion_locks WHERE expires > now()")->array->[0];
+  $stats{uptime} = $db->query( "SHOW GLOBAL STATUS LIKE 'Uptime'" )->hash->{Value};
+
+  return \%stats;
 }
 
 sub unregister_worker {
@@ -1153,7 +1160,7 @@ Minion::Backend::mysql
 
 =head1 VERSION
 
-version 0.27
+version 0.28
 
 =head1 SYNOPSIS
 
@@ -1805,3 +1812,6 @@ CREATE INDEX minion_notes_note_key ON minion_notes (note_key);
 -- 10 down
 DROP INDEX minion_notes_note_key ON minion_notes;
 
+-- 11 up
+CREATE INDEX minion_jobs_stats_idx ON minion_jobs (state, `delayed`);
+
